/******************************************************************************
* ToolBox 1.0 - Multi-Call Binary Maker (launcher)
* Copyright © JakeSFR 2019
* Zero-Clause BSD License applies
******************************************************************************/

#define VERSION "1.0"

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <libgen.h>
#include <errno.h>
#include <sys/syscall.h>
#include <linux/memfd.h>

#include "applets.h"	/* generated by toolbox.sh */

extern char **environ;

/* ------------------------------------------------------------------------- */

void usage(char *appname) {

	printf("Usage:\n");
	printf("%s %-30s%s\n", appname, "<applet> [arguments]", "execute an applet");
	printf("%s %-30s%s\n", appname, "-l|--list", "list available applets");
	printf("%s %-30s%s\n", appname, "-d|--dump <applet>", "dump an applet to stdout");
	printf("%s %-30s%s\n", appname, "-v|--version", "show version info");
	printf("%s %-30s%s\n", appname, "-h|--help", "this very help");

}

/* ------------------------------------------------------------------------- */

void version(void) {

	printf(
		"ToolBox-" VERSION " - Multi-Call Binary Maker\n"
		"Copyright © JakeSFR 2019\n"
		"Zero-Clause BSD License applies\n"
	);

}

/* ------------------------------------------------------------------------- */

int find_index(char *myname, char *appname) {
	
	int index;

	for (index = 0; index < sizeof(appnames) / sizeof(appnames[0]); index++) {
		if (strcmp(appname, appnames[index]) == 0) return index;
	}

	if(myname == NULL) {

		return -1;

	} else {

		fprintf(stderr, "%s: '%s' is not included\n", myname, appname);
		exit(EXIT_FAILURE);

	}

}

/* ------------------------------------------------------------------------- */

void app_list(void) {

	int i;

	for (i=0; i < sizeof(appnames) / sizeof(appnames[0]); i++) {
		if (strncmp(appnames[i], ".", 1) == 0) continue;	/* skip applet names that begin with . */
		printf("%s\n", appnames[i]);
	}

	exit(EXIT_SUCCESS);

}

/* ------------------------------------------------------------------------- */

void app_dump (char *myname, char *appname) {

	int index;
	
	index = find_index(myname, appname);

	if (strcmp(appnames[index], links_to[index]) == 0) {	/* not a link */

		fwrite(apps[index], sizeof(char), appsizes[index], stdout);
		fflush(stdout);
		exit(EXIT_SUCCESS);

	} else {												/* link */

		fprintf(stderr, "%s: it's a symlink, %s -> %s\n", myname, appnames[index], links_to[index]);
		fprintf(stdout, "%s", links_to[index]);
		fflush(stdout);
		exit(255);	/* so we know it's supposed to be a symlink */

	}

}

/* ------------------------------------------------------------------------- */

void app_exec (char *myname, int index, int argc, char *argv[]) {

	/* If MFD_CLOEXEC is set, shell scripts won't execute */
	/* int fd = syscall(SYS_memfd_create, appnames[index], MFD_CLOEXEC); */

	int fd = syscall(SYS_memfd_create, appnames[index], 0);

	if (fd < 0) {
		if (errno == ENOSYS) {
			fprintf(stderr, "%s: syscall 'memfd_create' is not available, aborting...\n", myname);
		} else {
			fprintf(stderr, "%s: failed to create file descriptor, aborting...\n", myname);
		}
		exit(EXIT_FAILURE);
	}

	if (! write(fd, apps[index], appsizes[index]) ) {
		fprintf(stderr, "%s: failed to write to file descriptor, aborting...\n", myname);
		exit(EXIT_FAILURE);
	}

	/* If the data in fd is an executable and everything went well,
	   no further instructions below this line will be executed */
	fexecve(fd, argv, environ);

	fprintf(stderr, "%s: unable to execute '%s'\n", myname, appnames[index]);
	
	exit(EXIT_FAILURE);

}

/* ------------------------------------------------------------------------- */

int main(int argc, char *argv[]) {

	int i, index;
	char *myname = argv[0];

	/* Handle Busybox-style symlinks */
	if ((index = find_index(NULL, basename(myname))) > -1)
		app_exec("toolbox", index, argc, argv);

	/* No args, show usage */
	if (argc == 1) {
		usage(myname);
		exit(EXIT_FAILURE);
	}

	/* Parse args */
	if (strcmp(argv[1], "-h") == 0 || strcmp(argv[1], "--help") == 0) {

		usage(myname);
		exit(EXIT_SUCCESS);

	} else if (strcmp(argv[1], "-v") == 0 || strcmp(argv[1], "--version") == 0) {

		version();
		exit(EXIT_SUCCESS);		
		
	} else if (strcmp(argv[1], "-l") == 0 || strcmp(argv[1], "--list") == 0) {

		app_list();

	} else if (strcmp(argv[1], "-d") == 0 || strcmp(argv[1], "--dump") == 0) {

		if (argc < 3) {
			fprintf(stderr, "%s: this option requires an argument\n", myname);
			exit(EXIT_FAILURE);
		}

		app_dump(myname, argv[2]);

	} else if (strncmp(argv[1], "-", 1) == 0) {

		fprintf(stderr, "%s: unrecognized option '%s'\n", myname, argv[1]);
		exit(EXIT_FAILURE);

	} else {

		index = find_index(myname, argv[1]);
		argc--;
		for (i = 0; i < argc; i++) argv[i] = argv[i+1];
		argv[i] = NULL;

		app_exec(myname, index, argc, argv);

	}

	exit(EXIT_SUCCESS);

}
