#!/bin/sh

###############################################################################
# ToolBox 1.0 - Multi-Call Binary Maker										  #
# Copyright © JakeSFR 2019													  #
# Zero-Clause BSD License applies		 									  #
###############################################################################

ARCH="${ARCH}"
CC="${CC:-gcc}"
LD="${LD:-ld}"
NM="${NM:-nm}"
AR="${AR:-ar}"
FLAGS="-Wall ${FLAGS}"

OUTNAME="${1:-toolbox}"
APPDIR="applets"

# -----------------------------------------------------------------------------

die() {
	[ "$@" ] && echo "$@"
	exit 1
}

# -----------------------------------------------------------------------------

set_arch() {
	
	# Check if ARCH is valid
	case "$ARCH" in
		i?86)	CC_ARCH="-m32"	LD_ARCH="-m elf_i386"		;;
		x86_64)	CC_ARCH="-m64"	LD_ARCH="-m elf_x86_64"		;;
		*)		CC_ARCH='' 		LD_ARCH=''					;;
	esac

}

# -----------------------------------------------------------------------------

check_memfd() {

	# Check if <linux/memfd.h> exists
	echo '#include <linux/memfd.h>
	int main(void){return 0;}' | ${CC} ${CC_ARCH} -o /dev/null -xc - || die

}

# -----------------------------------------------------------------------------

process_applets() {

	# Remove previously generated objects and initalize applets.h
	rm -f applets.a "${APPDIR}/"*.o "${APPDIR}/".*.o "${APPDIR}/".o
	: > applets.h
	OBJ=''
	
	APPNAMES="const char *appnames[] = {"
	APPSIZES="const unsigned long long int appsizes[] = {"
	APPS="const char *apps[] = {"
	LINKS_TO="const char *links_to[] = {"
	
	echo "Processing applets:"
	
	CNT=0
	for i in "${APPDIR}/".* "${APPDIR}/"* ; do
		
		# skip . and .. "files"
		[ "$i" = "${APPDIR}/." ] || [ "$i" = "${APPDIR}/.." ] || [ "$i" = "${APPDIR}/*" ] && continue
	
		# skip non-files
		if [ ! -f "$i" ]; then
			echo "WARNING: '${i##*/}' is not a file/symlink, skipping..."
			continue
		fi
	
		# skip symlinks pointing to files outside the 'applets' directory
		if [ -L "$i" ] && [ "$(basename "$(dirname "$(readlink -f "$i")")")" != "$APPDIR" ]; then
			echo "WARNING: '${i##*/}' is a symlink to a file outside '${APPDIR}', skipping..."
			continue
		fi
	
		APPNAME="$(echo "${i##*/}" | tr ' \t' '_')"						# actual name of an applet
		APPNAME_REAL="$(echo "$(basename "$(readlink -f "${i}")")")"	# the target, if APPNAME is a symlink
		APPNAME_REAL_NOSPC="$(echo "$APPNAME_REAL" | tr ' \t' '_')"
	
		if [ "$APPNAME" != "$APPNAME_REAL_NOSPC" ]; then
			echo "> ${APPNAME} -> ${APPNAME_REAL_NOSPC}"
		else
			echo "> ${APPNAME}"
		fi
	
		# Convert to an object, but only if an object doesn't exist yet
		if [ ! -f "${APPDIR}/${APPNAME_REAL_NOSPC}.o" ]; then
			${LD} ${LD_ARCH} -r -b binary "${APPDIR}/${APPNAME_REAL}" -o "${APPDIR}/${APPNAME_REAL_NOSPC}.o" || die
			OBJ="${OBJ} ${APPDIR}/${APPNAME_REAL_NOSPC}.o"
		fi
	
		# Find the _start symbol's name
		START="$(${NM} -P "${APPDIR}/${APPNAME_REAL_NOSPC}.o" | cut -f1 -d ' ' | grep '_start$')"
		[ "$APPNAME" = "$APPNAME_REAL_NOSPC" ] && echo "extern const char ${START};" >> applets.h
		
		# Get the size of an applet
		APPSIZE=$(stat -c %s "${APPDIR}/${APPNAME_REAL}")
	
		APPNAMES="${APPNAMES} \"${APPNAME}\","
		APPSIZES="${APPSIZES} ${APPSIZE},"
		APPS="${APPS} &${START},"
		LINKS_TO="${LINKS_TO} \"${APPNAME_REAL_NOSPC}\","
		
		CNT=$((CNT+1))
	
	done
	
	if [ $CNT -eq 0 ]; then
		die "No applets found in ${APPDIR} directory, aborting..."
	fi
	
	echo "${APPS%?} };"		>> applets.h
	echo "${APPNAMES%?} };" >> applets.h
	echo "${APPSIZES%?} };" >> applets.h
	echo "${LINKS_TO%?} };" >> applets.h
	
	ar q applets.a ${OBJ} || die

}

# -----------------------------------------------------------------------------

compile() {

	echo "Compiling, please wait..."
	
	${CC} ${CC_ARCH} toolbox.c applets.a -o "${OUTNAME}" ${FLAGS} || die

}

# -----------------------------------------------------------------------------

# START
set_arch
check_memfd
[ ! "$SKIP" ] && process_applets
compile

echo "Done!"

# =============================================================================
